unit Voxel3D;

interface

uses Voxel;

type
  RDT = Real; // data type to use in rotations etc
  TPoint3D = record
        X, Y, Z: RDT;
     end;
  TVoxel3DViewCell = record
        Colour: word;
        Normal: byte;
        Depth: RDT;
     end;
  TVoxel3D = class
     private
        Voxel: TVoxelSection; // owner
        Centre: TPoint3D;
        procedure CreateCanvas;
        procedure ClearCanvas;
     public
        Width, Height: Integer;
        Canvas: {packed} array of {packed} array of TVoxel3DViewCell;
        constructor Create(vxl: TVoxelSection);
        procedure Render(ax,ay,az: integer);
     end;

implementation

uses Math, SysUtils;

function Point(x,y,z: integer): TPoint3D;
begin
     Result.x := x;
     Result.y := y;
     Result.z := z;
end;

// precalculated sine and cosine tables for rotation optimisiation
var SINE,
    COSINE: array [0..360] of RDT;

procedure Rotate(p: TPoint3D; ax,ay,az: integer; var r: TPoint3D);
{ source: http://www.geocities.com/SiliconValley/Horizon/6933/3d.html
  x, y, z are the coordinates of the point to be rotated,
  ax, ay, az are the angles the object is to be rotated by around each axis,
  rx, ry, rz are the rotated coordinates
}
begin
     r.y := p.y*COSINE[ax] - p.z*SINE[ax]; // rotation around the x axis
     r.z := p.z*COSINE[ax] + p.y*SINE[ax];

     r.z := r.z*COSINE[ay] - p.x*SINE[ay]; // around y
     r.x := p.x*COSINE[ay] + r.z*SINE[ay];

     r.y := r.y*COSINE[az] + r.x*SINE[az];
     r.x := r.x*COSINE[az] - r.y*SINE[az]; // around z
end;

procedure Plot(p: TPoint3D; height: integer; var x,y: integer);
{ source: http://www.geocities.com/SiliconValley/Horizon/6933/3d.html }
var distance: integer;
begin
     distance := 1; // 256; //height;
     height := height div 2;
     if p.z = 0 then begin
        x := height;
        y := height;
     end else begin
        x := Round(height + distance*(p.x / p.z));
        y := Round(height - distance*(p.y / p.z));
     end;
end;

function Distance(p: TPoint3D): integer; // distance of p from (0,0,0)
begin
     Result := Ceil( SQRT (SQRT( SQR(p.x) + SQR(p.y)) + SQR(p.z) ));
end;

procedure TVoxel3D.CreateCanvas;
var x: integer;
begin
     // create canvas
     SetLength(Canvas,Width);
     for x := 0 to (Width - 1) do
         SetLength(Canvas[x],Height);
end;

procedure TVoxel3D.ClearCanvas;
var x, y: integer;
begin
     for x := 0 to (Width - 1) do
         for y := 0 to (Height - 1) do
             with Canvas[x,y] do begin
                  Colour := TRANSPARENT;
                  Normal := 0;
                  Depth := 0;
             end;
end;

constructor TVoxel3D.Create(vxl: TVoxelSection);
begin
     Voxel := vxl;
     // calculate width and height
     with vxl.Tailer do
          Width := Distance(Point(XSize,YSize,ZSize)) * 2;
     Height := Width; // square
     CreateCanvas;
     // calculate centre of model
     with Voxel.Tailer do begin
          Centre.X := XSize div 2;
          Centre.Y := YSize div 2;
          Centre.Z := ZSize div 2;

     end;
end;

procedure TVoxel3D.Render(ax,ay,az: integer);
var x, y, z,
    i, j: integer;
    v: TVoxelUnpacked;
    p,r: TPoint3D;
begin
     ClearCanvas;
     p.x := 0; // - Centre.x;
     for x := 0 to (voxel.Tailer.XSize - 1) do begin
         p.y := 0; // - Centre.y;
         for y := 0 to (voxel.Tailer.YSize - 1) do begin
             p.z := 0; // - Centre.z;
             for z := 0 to (voxel.Tailer.ZSize - 1) do begin
                 voxel.GetVoxel(x,y,z,v);
                 if v.Used then begin
                    Rotate(p,ax,ax,az,r);
                    Plot(r,Height,i,j);
                    Inc(i,Width div 2);
                    Inc(j,Height div 2);
                    if (i < 0) or (i >= Width) then Continue; // visible?
                    if (j < 0) or (j >= Height) then Continue; // visible?
                    with Canvas[i,j] do begin
                         if Depth > r.z then Continue; // in front of?
                         Colour := v.Colour;
                         Normal := v.Normal;
                         Depth := r.z;
                    end;
                 end;
                 p.z := p.z + 1;
             end;
             p.y := p.y + 1;
         end;
         p.x := p.x + 1;
     end;
end;

var deg: integer;
begin
     // init precalculated SIN/COSINE tables
     for deg := 0 to 360 do begin
         SINE[deg] := Sin(DegToRad(deg));
         COSINE[deg] := Cos(DegToRad(deg));
     end;
end.
