unit Voxel_Tools;

{ REVISION HISTORY
  20011218 removed all WriteLns
  20011213 using multi-pass flood filling instead of recursion
  20011213 fixed bug in Internal
}

interface

Uses Voxel;

// returns the number of redundant voxels removed
function RemoveRedundantVoxels(voxel: TVoxelSection): integer; // is more clever

// fills the Flags attribute of each voxel with facings information
// facings can be decoded by applying the above masks
// DOES NOT REQUIRE YOU TO REMOVE REDUNDANT VOXELS FIRST NOW!
procedure CalcFacings(voxel: TVoxelSection);

type TNormalsScheme = record
     author,
     FName: string;
     scheme: integer; // NOT_SET means ignore this whole scheme because there was an error loading it
     normal: array[0..255] of integer; // NOT_SET means not used
end;
type TApplyNormalsResult = record
     applied,
     redundant,
     confused: Integer;
end;
// loads a normals-scheme
function LoadNormalsScheme(filename: string): TNormalsScheme;
//applies a normals-scheme
function ApplyNormals(voxel: TVoxelSection; var NormalsScheme: TNormalsScheme): TApplyNormalsResult;

var NormalsSchemes: array of TNormalsScheme;
function LoadNormalsSchemes: Integer; // number loaded

const NOT_SET = -1;

function FacingAsStr(facing: integer; confused: boolean = false): string;

implementation

uses SysUtils, Classes;

const FACING_LEFT = 2;
      FACING_RIGHT = FACING_LEFT * 2;
      FACING_TOP = FACING_RIGHT * 2;
      FACING_BOTTOM = FACING_TOP * 2;
      FACING_FRONT = FACING_BOTTOM * 2;
      FACING_BACK = FACING_FRONT * 2;
      FACING_MASK = FACING_LEFT or FACING_RIGHT or FACING_TOP or FACING_BOTTOM or FACING_FRONT or FACING_BACK;

function BitSet(value, mask: integer): boolean;
begin
     Result := (value and mask) = mask;
end;

const isInside = 0;
      isOutside = 1;

// a voxel is internal if all adjacent voxels (including diagonals) are inside
function Internal(voxel: TVoxelSection; x,y,z: Integer): Boolean;
var maxx, maxy, maxz: Integer;
function IsExternal(x,y,z:Integer):Boolean;
var v: TVoxelUnpacked;
begin
     Result := True; // assume is
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     voxel.getVoxel(x,y,z,v);
     Result := BitSet(v.Flags,isOutside);
end;
var x1, y1, z1: Integer;
begin
     Result := False; // assume not
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     for x1 := Pred(x) to Succ(x) do
         for y1 := Pred(y) to Succ(y) do
             for z1 := Pred(z) to Succ(z) do
                 if IsExternal(x1,y1,z1) then
                    Exit;
     Result := True;
end;

function FindInside(voxel: TVoxelSection): Integer;
{ This function starts by assuming every cell is inside.
  It then does a flood fill from every face of the cube,
  flood-filling all empty cells to outside.
}
var maxx, maxy, maxz: integer;
procedure ClearFlags;
var v: TVoxelUnpacked;
    x, y, z: integer;
begin
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do begin
                 voxel.GetVoxel(x,y,z,v);
                 v.Flags := isInside;
                 voxel.SetVoxel(x,y,z,v);
             end;
end;
var OutsideCount: integer;
function IsOnOutside(x,y,z: Integer): Boolean;
procedure DoCheck(x,y,z: Integer);
var v: TVoxelUnpacked;
begin
     // bounds check!
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     // get the voxel
     voxel.GetVoxel(x,y,z,v);
     // is it outside already?
     if BitSet(v.Flags,isOutside) then
        Result := True;
end;
begin
     Result := False; // assume not
     DoCheck(x+1,y,z);
     DoCheck(x,y+1,z);
     DoCheck(x,y,z+1);
     DoCheck(x-1,y,z);
     DoCheck(x,y-1,z);
     DoCheck(x,y,z-1);
end;
var x, y, z, Pass, PassCount: integer;
    v: TVoxelUnpacked;
begin
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     ClearFlags;
     OutsideCount := 0;
     Pass := 0;
     repeat
          Inc(Pass);
          // use pass count for memory too
          PassCount := 0;
          // do a scan of all cells
          for x := 0 to maxx do
              for y := 0 to maxy do
                  for z := 0 to maxz do begin
                      voxel.GetVoxel(x,y,z,v);
                      // obvious checks
                      if v.Used then Continue; // used voxels are 'inside'
                      if BitSet(v.Flags,IsOutside) then Continue; // already found?
                      // check for extremities
                      if (x = 0) or (x = maxx) or
                         (y = 0) or (y = maxy) or
                         (z = 0) or (z = maxz) or
                      // or adjacent to an outside
                         IsOnOutside(x,y,z) then begin
                      // then set it to outside
                         v.Flags := isOutside;
                         voxel.SetVoxel(x,y,z,v);
                         Inc(PassCount);
                      end;
                  end;
          // keep running total
          Inc(OutsideCount,PassCount);
     until (Passcount = 0);
     Result := OutsideCount;
end;

procedure CalcFacings(voxel: TVoxelSection);
var maxx, maxy, maxz,
    x, y, z: integer;
procedure CalcFacing;
function Empty(x,y,z: integer): boolean;
var v: TVoxelUnpacked;
begin
     // check bounds
     Result := True; // outside the voxel is all empty
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     // ok, do the real check
     voxel.GetVoxel(x,y,z,v);
     if v.Used then
        Result := False
     else
         Result := BitSet(v.Flags,isOutside);
end;
function ChooseFromLeftOrRight(x,y,z: integer): integer; // returns the FACING to discard
var y1: Integer;
    v: TVoxelUnpacked;
    voidLeft, voidRight, nearestLeft: Boolean;
begin
     // void to left?
     voidLeft := True;
     for y1 := Pred(y) downto 0 do begin
         voxel.GetVoxel(x,y1,z,v);
         if not BitSet(v.Flags,isOutside) then begin
            voidLeft := False;
            Break;
         end;
     end;
     // void to right?
     voidRight := True;
     for y1 := Succ(y) to maxy do begin
         voxel.GetVoxel(x,y1,z,v);
         if not BitSet(v.Flags,isOutside) then begin
            voidRight := False;
            Break;
         end;
     end;
     // decide what side of half-way it is
     nearestLeft := (y <= (maxy div 2));
     // determine facing to discard
     if voidLeft and not voidRight then
        Result := FACING_RIGHT
     else if voidRight and not voidLeft then
          Result := FACING_LEFT
     else if nearestLeft then
        Result := FACING_RIGHT
     else
         Result := FACING_LEFT;
end;
function ChooseFrontOrBack(x,y,z: integer): integer; // returns the FACING to discard
var x1: Integer;
    v: TVoxelUnpacked;
    voidFront, voidBack, nearestFront: Boolean;
begin
     // void in front?
     voidFront := True;
     for x1 := Pred(x) downto 0 do begin
         voxel.GetVoxel(x1,y,z,v);
         if not BitSet(v.Flags,isOutside) then begin
            voidFront := False;
            Break;
         end;
     end;
     // void to back?
     voidBack := True;
     for x1 := Succ(x) to maxx do begin
         voxel.GetVoxel(x1,y,z,v);
         if not BitSet(v.Flags,isOutside) then begin
            voidBack := False;
            Break;
         end;
     end;
     // decide what side of half-way it is
     nearestFront := (x <= (maxx div 2));
     // determine facing to discard
     if voidFront and not voidBack then
        Result := FACING_BACK
     else if voidBack and not voidFront then
          Result := FACING_FRONT
     else if nearestFront then
        Result := FACING_BACK
     else
         Result := FACING_FRONT;
end;
function ChooseTopOrBottom(x,y,z: Integer): Integer; // returns the FACING to discard
var z1: Integer;
    v: TVoxelUnpacked;
    voidAbove, voidBelow, nearestUp: Boolean;
begin
     // search up
     voidAbove := True;
     for z1 := Pred(z) downto 0 do begin
         voxel.GetVoxel(x,y,z1,v);
         if not BitSet(v.Flags,isOutside) then begin
            voidAbove := False;
            Break;
         end;
     end;
     // search down
     voidBelow := True;
     for z1 := Succ(z) to maxz do begin
         voxel.GetVoxel(x,y,z1,v);
         if not BitSet(v.Flags,isOutside) then begin
            voidBelow := False;
            Break;
         end;
     end;
     // decide which side of half-way it is
     nearestUp := (z >= (maxz div 2));
     // decide which facing to keep
     if voidAbove then // aways most important, e.g. low aircraft wings
        Result := FACING_TOP
     else if voidBelow then
          Result := FACING_BOTTOM
     else if nearestUp then
          Result := FACING_TOP
     else
         Result := FACING_BOTTOM;
     // swap result, to get which to *discard*
     if Result = FACING_TOP then
        Result := FACING_BOTTOM
     else
         Result := FACING_TOP;
end;
var v: TVoxelUnpacked;
    Flags: byte;
begin
     // get the voxel in question
     voxel.GetVoxel(x,y,z,v);
     // skip empty ones
     if not v.Used then Exit;
     Flags := v.Flags;
     // facings
     if Empty(x-1,y,z) then Flags := (Flags or FACING_FRONT);
     if Empty(x+1,y,z) then Flags := (Flags or FACING_BACK);
     if Empty(x,y-1,z) then Flags := (Flags or FACING_LEFT);
     if Empty(x,y+1,z) then Flags := (Flags or FACING_RIGHT);
     if Empty(x,y,z-1) then Flags := (Flags or FACING_BOTTOM);
     if Empty(x,y,z+1) then Flags := (Flags or FACING_TOP);
     // cancel out double-sided facings
     if BitSet(Flags,FACING_FRONT or FACING_BACK) then Dec(Flags,ChooseTopOrBottom(x,y,z));
     if BitSet(Flags,FACING_LEFT or FACING_RIGHT) then Dec(Flags,FACING_RIGHT);
     if BitSet(Flags,FACING_TOP or FACING_BOTTOM) then Dec(Flags,FACING_BOTTOM);
     // save the result
     v.Flags := Flags;
     voxel.SetVoxel(x,y,z,v);
end;
begin
     FindInside(voxel);
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     // examine each used voxel, pass one for outside voxels
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do
                 CalcFacing;
end;

function FacingAsStr(facing: integer; confused: boolean): string;
begin
     Result := '';
     if BitSet(facing,FACING_LEFT) then Result := Concat(Result,' LEFT');
     if BitSet(facing,FACING_RIGHT) then Result := Concat(Result,' RIGHT');
     if BitSet(facing,FACING_TOP) then Result := Concat(Result,' TOP');
     if BitSet(facing,FACING_BOTTOM) then Result := Concat(Result,' BOTTOM');
     if BitSet(facing,FACING_FRONT) then Result := Concat(Result,' FRONT');
     if BitSet(facing,FACING_BACK) then Result := Concat(Result,' BACK');
     Result := Trim(Result);
     if confused and (Result = '') then Result := 'CONFUSED';
end;

// loads a normals-scheme
function LoadNormalsScheme(filename: string): TNormalsScheme;

var F: TStrings;
    Key, Value: String;
    Section: integer;

procedure Replace(var s: string; const oldchar, newchar: char);
var i: integer;
begin
     for i := 1 to Length(s) do
         if s[i] = oldchar then
            s[i] := newchar;
end;

function Split(const Line: integer; var Key: String; var Value: String): boolean; // returns false if a bad line
var equals: integer;
    s: String;
begin
     Result := False; // assume bad
     s := F[Line];
     Replace(s,#9,#32); // replace tabs with spaces
     s := Trim(s);
     equals := Pos('=',s);
     if equals <= 1 then Exit; // no separator or starts with a space?
     Key := Trim(Copy(s,1,Pred(equals)));
     Value := Trim(Copy(s,Succ(equals),100));
     Result := True; // ok
end;

function isSection(const Line: integer): boolean;
var s: String;
begin
     s := F[Line];
     Replace(s,#9,#32); // replace tabs with spaces
     s := Trim(s);
     Result := (Pos('[',s) = 1) and (Pos(']',s) = Length(s));
end;

function GetSectionName(const Line: integer): String;
var s: String;
begin
     s := F[Line];
     Replace(s,#9,#32); // replace tabs with spaces
     s := Trim(s);
     if (Pos('[',s) = 1) and (Pos(']',s) = Length(s)) then
        Result := Trim(Copy(s,2,Length(s)-2))
     else
         Result := '';
end;

function SetSection(const name: String): boolean; // return FALSE if section not found
var sec: integer;
begin
     for Sec := 0 to F.Count - 1 do begin
         Section := Sec;
         if UpperCase(GetSectionName(Sec)) = Trim(UpperCase(name)) then begin
            Result := True;
            Exit;
         end;
     end;
     Result := False;
end;

function GetAsString(const name: String; const def: String): String; // return default if key not found
var Line: integer;
begin
     Result := def;
     Line := Succ(Section);
     while Line < F.Count do begin
           if IsSection(Line) then
              Break;
           Split(Line,Key,Value);
           if UpperCase(Key) = UpperCase(Trim(name)) then begin
              Result := Value;
              Break;
           end;
           Inc(Line);
     end;
end;

function GetAsInt(const name: String; const def: integer): integer; // return default if key not found
var Line: integer;
begin
     Result := def;
     try
        Line := Succ(Section);
        while Line < F.Count do begin
           if IsSection(Line) then
              Break;
           Split(Line,Key,Value);
           if UpperCase(Key) = UpperCase(Trim(name)) then begin
              Result := StrToInt(Value);
              Break;
           end;
           Inc(Line);
        end;
     except
        on EConvertError do
           Result := def;
     end;
end;

procedure Clear;
var normal: integer;
begin
     Result.FName := filename;
     Result.author := 'anonymous';
     Result.scheme := NOT_SET; // indicates bad
     for normal := 0 to 255 do
         Result.normal[normal] := NOT_SET; // indicates not assigned
end;
var Scheme, uKey: String;
    Facing: integer;
    Line, Normal: integer;
    E: Exception;
begin
     try
          Clear;
          if not FileExists(filename) then begin
             Exit;
          end;
          // open file
          F := TStringList.Create;
          F.LoadFromFile(filename);
          // get details
          if not SetSection('details') then
             raise Exception.Create('details section not found');
          Result.author := GetAsString('author','anonymous');
          Scheme := GetAsString('scheme','');
          if UpperCase(Scheme) = 'TIBERIAN_SUN' then
             Result.scheme := 02
          else if UpperCase(Scheme) = 'RED_ALERT_2' then
               Result.scheme := 04
          else if Scheme = '' then
               raise Exception.Create('scheme not specified')
          else
              raise Exception.Create(Concat('scheme "',Scheme,'" not valid'));
          // load normals
          if not SetSection('normals') then
             raise Exception.Create('normals section not found');
          for Line := Succ(Section) to F.Count - 1 do begin
              if isSection(Line) then Break; // end of section?
              if not Split(Line,Key,Value) then
                 Continue; // skip empty lines
              // get value
              try
                 Normal := StrToInt(Value);
              except
                    on EConvertError do
                       Normal := NOT_SET;
              end;
              if Normal > NOT_SET then // out of range checks
                 if Result.scheme = 02 then begin // TS
                    if Normal > 35 then Normal := NOT_SET;
                 end else if Normal > 255 then // RA2
                     Normal := NOT_SET;
              if Normal <= NOT_SET then
                 Normal := NOT_SET
              else begin
                  // determine mask
                  Facing := 0;
                  uKey := UpperCase(Key);
                  if Pos('LEFT',uKey) > 0 then Facing := Facing or FACING_LEFT;
                  if Pos('RIGHT',uKey) > 0 then Facing := Facing or FACING_RIGHT;
                  if Pos('TOP',uKey) > 0 then Facing := Facing or FACING_TOP;
                  if Pos('BOTTOM',uKey) > 0 then Facing := Facing or FACING_BOTTOM;
                  if Pos('FRONT',uKey) > 0 then Facing := Facing or FACING_FRONT;
                  if Pos('BACK',uKey) > 0 then Facing := Facing or FACING_BACK;
                  // save it
                  Result.normal[Facing] := Normal;
              end;
          end;
          // done
          F.Free;
     except
           on E: Exception do
              Result.scheme := NOT_SET;
     end;
end;

//applies a normals-scheme
function ApplyNormals(voxel: TVoxelSection; var NormalsScheme: TNormalsScheme): TApplyNormalsResult;
var maxx, maxy, maxz,
    x, y, z: integer;
procedure ApplyNormal;
var v: TVoxelUnpacked;
    Facing, Normal: integer;
begin
     voxel.GetVoxel(x,y,z,v);
     if v.Used then begin
        if Internal(voxel,x,y,z) then begin
           Inc(Result.redundant);
           Exit;
        end;
        Facing := v.Flags and FACING_MASK;
        Normal := NormalsScheme.normal[Facing];
        if Normal <> NOT_SET then begin
           v.Normal := Normal;
           voxel.SetVoxel(x,y,z,v);
           Inc(Result.applied);
        end else
            Inc(Result.confused);
     end;
end;
begin
     // constraints
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     // set right mode
     voxel.Tailer.Unknown := NormalsScheme.scheme;
     // find the facings
     CalcFacings(voxel);
     // examine each used voxel.
     with Result do begin
        applied := 0;
        redundant := 0;
        confused := 0;
     end;
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do
                 ApplyNormal;
end;

function RemoveRedundantVoxels(voxel: TVoxelSection): integer;
var x,y,z,
    maxx, maxy, maxz: Integer;
    v: TVoxelUnpacked;
begin
     Result := 0;
     // first, make sure flags are ok
     FindInside(voxel);
     // constraints
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     // now test it
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do
                 if Internal(voxel,x,y,z) then begin
                    voxel.GetVoxel(x,y,z,v);
                    if v.Used then begin
                       v.Used := False;
                       voxel.SetVoxel(x,y,z,v);
                       Inc(Result);
                    end;
                 end;
end;

function LoadNormalsSchemes: Integer; // number loaded
var     f: TSearchRec;
        found: Integer;
        path: String;
        scheme: TNormalsScheme;
begin
        // prepare
        Result := 0;
        SetLength(NormalsSchemes,Result);
        path := Concat(ExtractFilePath(ParamStr(0)),'\*.normals');
        // find files
        if FindFirst(path,faAnyFile,f) = 0 then repeat
                scheme := LoadNormalsScheme(f.Name);
                if scheme.scheme <> NOT_SET then begin // load ok?
                        SetLength(NormalsSchemes,Succ(Result));
                        NormalsSchemes[Result] := scheme;
                        Inc(Result);
                end;
        until FindNext(f) <> 0;
        FindClose(f);
end;

end.

